/* Nessus
 * Copyright (C) 1998, 1999, 2000 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Attack.c : 
 *  launch the plugins, and manages the multithreading 
 *
 */
 
#include <includes.h>
#include "attack.h"
#include "log.h"
#include "hosts_gatherer.h"
#include "sighand.h"
#include "rules.h"
#include "auth.h"
#include "threads.h"
#include "comm.h" 
#include "utils.h"
#include "preferences.h"
#include "plugs_deps.h"
#include "ntp.h"
#include "ntp_11.h"
#include "pluginload.h"
#include "plugs_req.h"

#ifdef ENABLE_CRYPTO_LAYER
#include "peks/peks.h"
#endif

extern int Delay_between_tests;
extern u_short * getpts(char *);

/*******************************************************

		PRIVATE FUNCTIONS
		
********************************************************/
/*-------------------------------------------------------
		
		Determine if our father
		process is alive

 -------------------------------------------------------*/
 
static int
is_father_alive(globals)
 struct arglist * globals;
{
 pid_t pid = (pid_t)arg_get_value(globals, "father");
 if(pid)
  return kill(pid, 0);
 else
  return 0;
}


/*-------------------------------------------------------

	Init. an arglist which can be used by 
	the plugins from an hostname and its ip
	
--------------------------------------------------------*/	
static struct arglist * 
attack_init_hostinfos(hostname, ip)
     char * hostname;
    struct in_addr * ip;
{
  struct arglist * hostinfos;
  struct arglist * ports;
  
  hostinfos = emalloc(sizeof(struct arglist));
  arg_add_value(hostinfos, "FQDN", ARG_STRING, strlen(hostname), hostname);
  arg_add_value(hostinfos, "IP", ARG_PTR, sizeof(struct in_addr), ip);
  ports = emalloc(sizeof(struct arglist));
  arg_add_value(hostinfos, "PORTS", ARG_ARGLIST, sizeof(struct arglist), ports);
  return(hostinfos);
}





/*--------------------------------------------------------
	
		 Return our user name
 
 ---------------------------------------------------------*/
 
static char *
attack_user_name(globals)
 struct arglist * globals;
{
 return (char *)arg_get_value(globals, "user");
}
	

/*---------------------------------------------------------

	Wait if old and new have some port in common
	
----------------------------------------------------------*/
static int
attack_wait_if_similar_plugins(globals, old, new, preferences)
 struct arglist * globals, *old, *new, *preferences;
{
 if(old &&  
    Delay_between_tests &&
    requirements_common_ports(new->value, 
    			  old->value))
    {
    if(preferences_log_whole_attack(preferences))
    log_write("user %s : Two consecutive attacks on the same port. Sleeping %d second%s",
	  		attack_user_name(globals),
			Delay_between_tests, 
			(Delay_between_tests==1) ?"":"s");    
    sleep(Delay_between_tests);
  }
  return 0;
}



/*--------------------------------------------------------
	
	          Attack _one_ host

----------------------------------------------------------*/	
static void 
attack_host(globals, 
		  hostinfos, 
		  plugins, 
		  hostname, 
		  port_range, 
		  hosts_list)
     
     
     struct arglist * globals;
     struct arglist * hostinfos;
     struct arglist * plugins;
     char * hostname;
     char * port_range;
     struct arglist * hosts_list;
{ 

  /*
   * Used for the status
   */
  int num_plugs = 0;
  int cur_plug = 1;

 
  struct arglist * previous_plugin = NULL;
  struct arglist * preferences = arg_get_value(globals,"preferences");
  struct arglist * key = emalloc(sizeof(struct arglist));
  struct arglist * portsinfos = emalloc(sizeof(struct arglist));
  int optimize = preferences_optimize_test(preferences);;
  
  
  
  
  arg_add_value(key, "hosts", ARG_ARGLIST, -1, hosts_list);
  num_plugs = get_active_plugins_number(plugins);
 
  arg_add_value(portsinfos, "ports", ARG_PTR, -1, arg_get_value(globals,"ports"));
  arg_add_value(portsinfos, "ports_num", ARG_INT, -1, arg_get_value(globals,
  								"ports_num"));
  arg_add_value(portsinfos, "key", ARG_ARGLIST, -1, key);
  /* launch the plugins */
  
  while(plugins && plugins->next)
    {
      char * name = (char*)arg_get_value(plugins->value, "full_name");
  
      if(is_father_alive(globals))return;
      if(plug_get_launch(plugins->value)) /* can we launch it ? */
	{
	 char * error;
	 pl_class_t * cl_ptr = arg_get_value(plugins->value, "PLUGIN_CLASS");
	 comm_send_status(globals, hostname, "attack", cur_plug++, num_plugs);
		   
	 if(!optimize || 
	   !(error = requirements_plugin(key, plugins,portsinfos)))
	 {
	 if(preferences_log_whole_attack(preferences))
	 	log_write("user %s : launching %s against %s\n", 
	 				attack_user_name(globals),
					plugins->name, 
					hostname);
					
 	 attack_wait_if_similar_plugins(globals,
	 				previous_plugin, 
	 				plugins, 
					preferences); 
	 /*
	  * Start the plugin
	  */
	 (*cl_ptr->pl_launch)(globals, 
			     plugins->value, 
			     hostinfos,
			     preferences,
			     key,
			     name);
			     
	 /*
	  * Stop the test if the host is 'dead'
	  */	 
        if(arg_get_value(key, "Host/dead"))
	{
	  log_write("user %s : The remote host (%s) is dead\n",
	  				attack_user_name(globals),
	  				hostname);
	  return;
	}
	
        previous_plugin = plugins; 
       }
       
       else /* requirements_plugin() failed */
	  {
	   if(preferences_log_whole_attack(preferences))
	    log_write("user %s : Not launching %s against %s %s (this is not an error)\n",
	       			attack_user_name(globals),
				plugins->name, 
				hostname, 
				error);
	
	    free(error);	
	  }
      } /* if(plugins->launch) */
      
      plugins = plugins->next;
      
    } /* while(plugins) */
    
    free(portsinfos);
}

/*-----------------------------------------------------------------

  attack_start : set up some data and jump into
  attack_host()

 -----------------------------------------------------------------*/
static void
attack_start(args)
  struct arglist * args;
{
 struct linger linger;
 struct arglist * globals = arg_get_value(args, "globals");
 char * hostname = arg_get_value(args, "hostname");
 struct arglist * plugs = arg_get_value(args, "plugins");
 char * port_range = arg_get_value(args, "port_range");
 struct in_addr * hostip = arg_get_value(args, "hostip");
 struct arglist * hosts_list = arg_get_value(args, "hosts_list");
 struct arglist * hostinfos;
 
 struct arglist * preferences = arg_get_value(globals,"preferences");
 int thread_socket = (int)arg_get_value(globals, "thread_socket");
 int soc;
 struct arglist * new_plugs = NULL;
 

 /*
  * Options regarding the communication with out father
  */
 linger.l_onoff = 1;
 linger.l_linger = 60;
 setsockopt(thread_socket, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger));
 close((int)arg_get_value(globals, "global_socket"));
 arg_set_value(globals, "global_socket", -1, (void*)thread_socket);
 
 /*
  * Wait for the server to confirm it read our data
  * (prevents client desynch)
  */
 arg_add_value(globals, "confirm", ARG_INT, sizeof(int), (void*)1);
 
 soc = thread_socket;

 
 

   hostinfos = attack_init_hostinfos(hostname,hostip);
  
 #ifndef USE_FORK_THREADS
    new_plugs = emalloc(sizeof(struct arglist));
    arg_dup(new_plugs, plugs);
#else
    new_plugs = plugs;
#endif
    
  plugins_set_socket(plugs, soc);
  attack_host(globals, hostinfos, new_plugs, hostname, port_range, hosts_list);
    
  if(preferences_ntp_show_end(preferences))ntp_11_show_end(globals, hostname);
     
  shutdown(soc, 2);
  close(soc);
}

/*******************************************************

		PUBLIC FUNCTIONS
		
********************************************************/


/*------------------------------------------------

   This function attacks a whole network
 		
 -----------------------------------------------*/
int 
attack_network(globals)
    struct arglist * globals;
{
  int max_threads;
  int num_tested = 0;
  int host_pending;
  int num_threads = 0;
  char * hostname;
  struct in_addr host_ip;
  char * port_range;
  int hg_flags;
  struct hg_globals * hg_globals;
  int global_socket = (int)arg_get_value(globals, "global_socket");
  struct arglist * preferences = arg_get_value(globals, "preferences");
  struct arglist * plugins = arg_get_value(globals, "plugins");
  struct arglist * hosts_list = emalloc(sizeof(struct arglist));
  struct attack_atom ** atoms = attack_atom_new();
  ntp_caps* caps = arg_get_value(globals, "ntp_caps");
  struct nessusd_threads ** threads;
  struct nessus_rules *rules = arg_get_value(globals, "rules");
  struct arglist * tested_hosts = 0;
  struct arglist * rejected_hosts = emalloc(sizeof(struct arglist));
  char * t;
  unsigned short * ports;
  int data;
 
  hostname = arg_get_value(preferences, "TARGET");
  port_range = arg_get_value(preferences, "port_range");
  if(!port_range)port_range = "1-15000";
  if(port_range && strlen(port_range) && strcmp(port_range, "-1"))
  {
   int num = 0;
   int i = 0;
   ports = (unsigned short*)getpts(port_range);
   for(i=0;ports[i];i++,num++);
   if(arg_get_value(globals, "ports"))
    arg_set_value(globals, "ports", -1, ports);
   else
    arg_add_value(globals, "ports", ARG_PTR,-1, ports);
    
   if(arg_get_value(globals, "ports_num"))
    arg_set_value(globals, "ports_num", sizeof(int),  (void*)num);
   else
    arg_add_value(globals, "ports_num", ARG_INT, sizeof(int), (void*)num);
  }
  /*
   * Initialization of the attack
   */
   
  
  if(caps->ntp_11) send_plugin_order(globals, plugins);
  hg_flags = preferences_get_host_expansion(preferences);

  max_threads = get_max_thread_number(preferences);
  log_write("user %s starts a new attack. Target(s) : %s, with max_threads = %d\n",
			 attack_user_name(globals), 
			 hostname,
			 max_threads);
			 
  /* 
   * Initialize the hosts_gatherer library 
   */
   			  
  hg_globals = hg_init(hostname, hg_flags); 
  hostname = hg_next_host(hg_globals,&host_ip);
  
  if(hostname)
  {

   if(preferences_log_whole_attack(preferences))
    log_write("user %s : testing %s [%s]\n",
			 attack_user_name(globals), 
			 hostname,
			 inet_ntoa(host_ip));
   threads = nessusd_threads_init();
   if(arg_get_value(globals, "threads"))
    arg_set_value(globals, "threads", -1, threads);
   else
    arg_add_value(globals, "threads", ARG_PTR, -1, threads);
  
   tested_hosts = emalloc(sizeof(struct arglist));
  /*
   * Start the attack !
   */
 
   arg_add_value(globals, "client_socket", ARG_INT, sizeof(int), (void*)global_socket);       
   arg_add_value(globals, "father", ARG_INT, sizeof(int), (void*)getpid());
   while(hostname)
    {
      nthread_t pid;
 
      host_pending = 0 ;
      if(CAN_TEST(get_host_rules(rules, host_ip,32))) /* do we have the right to test this host ? */ 
      {
          t = estrdup(hostname);
  	  arg_add_value(tested_hosts, "", ARG_STRING, strlen(t), t);
	  if(num_threads < max_threads)
	    {
             struct arglist * arglist = emalloc(sizeof(struct arglist));
	     struct in_addr * ia = emalloc(sizeof(ia));
	     int s[2];
	     
	     if(!socketpair(AF_UNIX, SOCK_STREAM, 0, s))
	     {
	      if(arg_get_value(globals, "thread_socket"))
	       arg_set_value(globals, "thread_socket", sizeof(int), (void*)s[0]);
	      else
	       arg_add_value(globals, "thread_socket", ARG_INT, sizeof(int), (void*)s[0]);
	   
	      attack_atom_insert(atoms, hostname, s[1], s[0]);
	      }
	     else perror("socketpair ");	
	     arg_add_value(hosts_list, t, ARG_INT, -1, (void *)1);
             arg_add_value(arglist, "globals", ARG_ARGLIST, -1, globals);
             arg_add_value(arglist, "hostname", ARG_STRING, -1, t);
	     ia->s_addr = host_ip.s_addr;
	     arg_add_value(arglist, "hostip", ARG_PTR, -1, ia); 
             arg_add_value(arglist, "plugins", ARG_ARGLIST, -1, plugins);
             arg_add_value(arglist, "port_range", ARG_STRING, -1, port_range);
             arg_add_value(arglist, "hosts_list", ARG_ARGLIST, -1, hosts_list); 
	     pid = create_thread(attack_start, arglist, (struct arglist*)-1);   
	     nessusd_thread_register(threads, t, pid);
             num_threads++;
	     log_write("user %s : testing %s [%d]\n", attack_user_name(globals), hostname, pid);
	     arg_free(arglist);
	    }
	  else 
	    {
              /* 
               * There are too many threads at this time... We
               * have to wait until one of them finishes
               */
	      host_pending = 1; /* means that we must NOT update the hostname */
             
	      while(num_threads >= max_threads)
                {
		/*
		 * Check the threads input
		 */
		if(check_threads_input(atoms, global_socket, globals)<0)
			goto stop;
		num_threads = wait_for_a_thread(threads);
                }
	    }
	} else {
		log_write("user %s : rejected attempt to scan %s", 
			attack_user_name(globals), hostname);
		  arg_add_value(rejected_hosts, hostname, ARG_INT, sizeof(int), (void*)1);	
		}	
      if(!host_pending)
      {
        num_tested++;
	free(hostname);
       	hostname = hg_next_host(hg_globals, &host_ip);
      }
     }
 
   
    
 
  /*
   * Every host is being tested... We have to wait for the threads
   * to terminate
   */
  
  
  while(wait_for_a_thread(threads)){
	/*
	 * check threads input
	 */
	int n = check_threads_input(atoms, global_socket, globals);
	while(n)
	 {
	 n = check_threads_input(atoms, global_socket, globals);
	 if(n < 0)
	  return -1;
	 wait_for_a_thread(threads);
	 }  
        }
  	
  
  /*
   * XXX
   * At this very moment, external programs
   * may be called using the string 
   * created by the function : hosts_arglist_to_string(tested_hosts)
   * which returns the list of hosts that shall be tested
   */
   
   data = check_threads_input(atoms, global_socket, globals);
   while(data>0)
     data = check_threads_input(atoms, global_socket, globals);
      
      
    
    
    
   /*
    * some_function(hosts_arglist_to_string(tested_hosts));
    */
    
   log_write("user %s : test complete", attack_user_name(globals));
 
  }
 
  if(rejected_hosts && rejected_hosts->next)
   {
     char * banner = emalloc(4001);
     int length = 0;

     sprintf(banner, "SERVER <|> ERROR <|> These hosts could not be tested because you are not allowed to do so :;");
     length = strlen(banner);

     while(rejected_hosts->next && (length < (4000-3)))
	{
	  int n;
	  n = strlen(rejected_hosts->name);
	  if(length + n + 1 >= 4000)
	  {
	    n = 4000 - length  - 2;
	  }
	  strncat(banner, rejected_hosts->name, n);
	  strncat(banner, ";", 1);
	  length+=n+1;
	  rejected_hosts = rejected_hosts->next;
	}
      if(rejected_hosts->next)
       strcat(banner, "...");
       
     auth_printf(globals, "%s\n", banner);
   }
	
stop:
  hg_cleanup(hg_globals);
  arg_free(tested_hosts);
  return(0);
}

   
 
